import asyncio
from abc import abstractmethod
import logging
import json
from .const import *
logger = logging.getLogger("Airzone")
logger.setLevel(logging.INFO) 

class AirzoneDevice:
    def __init__(self, device_id, ws_id, conn, installation, name=None):
        self._conn = conn
        self._installation = installation
        self._id = device_id
        self._name = name
        self._ws_id = ws_id
        self.status = {}
        self.adv_conf = {}
        self.sched_conf = {}
        self.user_conf = {}
        self._callbacks = []
        if getattr(self, "ws_command_listeners", None) is None:
            self.ws_command_listeners = []
        self.ws_command_listeners.append(
            {"command": f"DEVICES_UPDATES.{ws_id}.{self._id}", "method": self.update}
        )
        self.ws_command_listeners.append(
            {"command": "DEVICE_STATE", "method": self.initialize}
        )

        for listener in self.ws_command_listeners:
            conn.get_websocket().add_command_callback(
                listener["command"], listener["method"]
            )

    def add_change_callback(self, callback):
        self._callbacks.append(callback)

    @staticmethod
    def get_device(data, conn, installation):
        _class = globals()["AirzoneDevice_" + data["type"]]
        device = _class(
            data["device_id"],
            data["ws_id"],
            conn,
            installation,
            data.get("name") or ("Sin nombre " + data["type"]),
        )
        return device

    def get_installation(self):
        return self._installation

    async def update(self, command_no, message_data, ws_client):
        try:
            for setting in message_data["change"]:
                for key, value in message_data["change"][setting].items():
                    self.__getattribute__(setting)[key] = value

        except:
            logger.exception(json.dumps(message_data))
        for callback in self._callbacks:
            callback()

    async def initialize(self, command_no, message_data, ws_client):
        if message_data["device_id"] == self._id:

            self.status = message_data["status"]
            self.adv_conf = message_data.get("adv_conf", {})

    @abstractmethod
    async def get_capabilities(self):
        pass


class AirzoneDevice_az_ccp(AirzoneDevice):
    pass


class AirzoneDevice_az_system(AirzoneDevice):
    def __init__(self, device_id, ws_id, conn, installation, name=None):
        self.ws_command_listeners=[]
        super().__init__(device_id, ws_id, conn, installation, name)

    async def get_capabilities(self):
        while self.status == {}:
            await asyncio.sleep(0.1)
        return self.status

    def get_system_status(self):
        return self.get_installation().get_system_status()
    
    def get_system_status(self):
        return self.get_installation().get_system_idle()


    async def set_mode(self, mode):
        await self._conn.patch(
            f"/devices/{self._id}",
            {
                "installation_id": self._installation._id,
                "opts": {"units": CONST_CELSIUS},
                "param": "mode",
                "value": mode,
            },
        )

    async def set_power(self, power):
        self.status["power"] = power
        await self._conn.patch(
            f"/devices/{self._id}",
            {
                "installation_id": self._installation._id,
                "opts": {"units": CONST_CELSIUS},
                "param": "power",
                "value": power,
            },
        )


class AirzoneDevice_az_zone(AirzoneDevice):
    async def get_capabilities(self):
        while self.status == {}:
            await asyncio.sleep(0.1)
        return self.status

    async def set_temperature(self, temperature: float):
        await self._conn.patch(
            f"/devices/{self._id}",
            {
                "installation_id": self._installation._id,
                "opts": {"units": CONST_CELSIUS},
                "param": "setpoint_air_heat",
                "value": temperature,
            },
        )

    async def set_mode(self, mode):
        await self._conn.patch(
            f"/devices/{self._id}",
            {
                "installation_id": self._installation._id,
                "opts": {"units": CONST_CELSIUS},
                "param": "mode",
                "value": mode,
            },
        )

    async def set_power(self, power):
        self.status["power"] = power
        await self._conn.patch(
            f"/devices/{self._id}",
            {
                "installation_id": self._installation._id,
                "opts": {"units": CONST_CELSIUS},
                "param": "power",
                "value": power,
            },
        )



CONST_CELSIUS = 0
CONST_FARENHEIT = 1

# Websocket opcodes
OP_CODE_PING = "2"
OP_CODE_PONG = "3"
OP_CODE_CONNECTED = "0"
OP_CODE_CONNECTED_ACK = "40"
OP_CODE_COMMAND = "42"
OP_CODE_COMMAND_RESPONSE = "43"



# import asyncio
# from unittest import mock
# import pytest
# from unittest.mock import AsyncMock, MagicMock

# from airzonecloud.AirzoneDevice import *
# from airzonecloud.AirzoneWebsocket import AirzoneWebsocketCLient
# from airzonecloud.AirzoneCloudConnector import AirzoneCloudConnector

# import logging

# logging.basicConfig(level=logging.DEBUG)
# mylogger = logging.getLogger()

# def test_get_device_return_correct_class(mocker):
#     with AsyncMock() as conn_mock:
#         device = AirzoneDevice.get_device({"device_id":1,"type":"az_zone","ws_id":1},conn_mock,None)
#         assert isinstance(device,AirzoneDevice_az_zone)
#         device = AirzoneDevice.get_device({"device_id":1,"type":"az_system","ws_id":1},conn_mock,None)
#         assert isinstance(device,AirzoneDevice_az_system)
#         device = AirzoneDevice.get_device({"device_id":1,"type":"az_ccp","ws_id":1},conn_mock,None)
#         assert isinstance(device,AirzoneDevice_az_ccp)
#         
# def test_get_device_assigns_name_to_new_device_from_data(mocker):
#     with AsyncMock() as conn_mock:
#         device = AirzoneDevice.get_device({"device_id":1,"type":"az_zone","ws_id":1,"name":"new_name"},conn_mock,None)
#         assert device._name == "new_name"

# def test_get_device_assigns_name_to_new_device_if_no_name_in_data(mocker):
#     with AsyncMock() as conn_mock:
#         device = AirzoneDevice.get_device({"device_id":1,"type":"az_zone","ws_id":1},conn_mock,None)
#         assert device._name != ""
#         assert device._name == "Sin nombre az_zone"

# def test_device_appends_listener_to_command_update_on_instantation(mocker): 
#     # prepare a fake connection
#     conn = AirzoneCloudConnector("user","pass")
#     ws_id="ws1"
#     _id="id1"
#     device = AirzoneDevice(_id,ws_id,conn,None)
#     
#     assert  f"DEVICES_UPDATES.{ws_id}.{_id}" in [key for key in  conn.get_websocket()._command_callbacks]

# @pytest.mark.asyncio
# async def test_device_update_listener_is_called_on_right_message(mocker):
#     # prepare a fake connection
#     update_patch = mocker.patch('airzonecloud.AirzoneDevice.AirzoneDevice.update')
#     conn = AirzoneCloudConnector("user","pass")
#     
#     ws_id="ws1"
#     _id="id1"
#     device = AirzoneDevice(_id,ws_id,conn,None)
#     command=[ f"DEVICES_UPDATES.{ws_id}.{_id}" ,{"hello":1}]
#     mylogger.info(command)
#     conn.get_websocket().process_request("42"+json.dumps(command),None)
#     update_patch.assert_awaited_with({"hello":1})
#     




import aiohttp
import asyncio
from .AirzoneWebsocket import AirzoneWebsocketCLient



class AirzoneCloudConnector:
    def __init__(
        self,
        user,
        password,
        host="m.airzonecloud.com",
        protocol="https",
        api_path="/api/v1",
    ):
        self._user = user
        self._password = password
        self._host = host
        self._protocol = protocol
        self._api_path = api_path
        self._connected = False
        self._token = None
        self._refresh_token = None
        self._base_url = f"{self._protocol}://{self._host}{self._api_path}"
        self.ws_client = AirzoneWebsocketCLient()

        self.ws_client.append_startup_command(self.ws_client.clear_listeners)
        self.ws_client.append_startup_command(self.ws_client.clear_listeners)

    def get_websocket(self):
        return self.ws_client

    async def get(self, method, query_params={}, headers={}, authenticate=True):
        if authenticate:
            if not self._connected:
                raise Exception("Not connected")
            headers["authorization"] = f"Bearer {self._token}"
        async with aiohttp.ClientSession() as session:
            async with session.get(
                self._base_url + method, headers=headers
            ) as response:
                if "application/json" in response.headers.get("content-type", ""):
                    return await response.json()
                return await response.text()

    async def post(self, method, body={}, headers={}, authenticate=True):
        if authenticate:
            if not self._connected:
                raise Exception("Not connected")
            headers["authorization"] = f"Bearer {self._token}"
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self._base_url + method, json=body, headers=headers
            ) as response:
                if "application/json" in response.headers.get("content-type", ""):
                    return await response.json()
                return await response.text()

    async def patch(self, method, body=None, headers=None, authenticate=True):
        if headers is None:
            headers = {}

        if authenticate:
            if not self._connected:
                raise Exception("Not connected")
            headers["authorization"] = f"Bearer {self._token}"
        async with aiohttp.ClientSession() as session:
            async with session.patch(
                self._base_url + method, json=body, headers=headers
            ) as response:
                if "application/json" in response.headers.get("content-type", ""):
                    return await response.json()
                return await response.text()

    async def connect(self):
        if self._connected:
            return True
        response = await self.post(
            "/auth/login",
            {"email": self._user, "password": self._password},
            authenticate=False,
        )
        self._token = response["token"]
        self._refresh_token = response["refreshToken"]
        self._connected = True
        self.ws_client.set_token(self._token, self._refresh_token)
        return True

    async def refresh_connection(self):
        response = await self.get(f"/auth/refreshToken/{self._refresh_token}")
        self._token = response["token"]
        self._refresh_token = response["refreshToken"]
        self._connected = True
        return True

    async def get_installations(self):
        if not self._connected:
            await self.connect()
        response = await self.get("/installations")
        return response["installations"]

    async def get_installation(self, installation_id):
        if not self._connected:
            await self.connect()
        return await self.get(f"/installations/{installation_id}")

    async def get_location(self, location_id):
        if not self._connected:
            await self.connect()
        return await self.get(f"/installations/location/{location_id}")



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="python-airzonecloud",
    version="0.0.8",
    author="José Luis Maroto",
    author_email="maroto.joseluis@gmail.com",
    description="A package to connect to AirZone cloud",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jlmaroto/python-airzonecloud",
    packages=["airzonecloud"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import asyncio
import pytest
from airzonecloud.AirzoneDevice import AirzoneDevice_az_system

from airzonecloud.AirzoneInstallation import AirzoneInstallation, AirzoneCloudConnector
from unittest.mock import AsyncMock, MagicMock

def test_if_provided_connection_uses_that_connection():
    conn = object()
    installation = AirzoneInstallation("1",conn=conn, email="j@j.com",password="1234")
    assert conn == installation._conn
    
def test_if_not_connexion_create_one():
    installation = AirzoneInstallation("1",email="j@j.com",password="1234")
    assert isinstance(installation._conn,AirzoneCloudConnector)

@pytest.mark.asyncio
async def test_installation_start_get_installation_info(mocker):
    get_installation_patch = mocker.patch('airzonecloud.AirzoneCloudConnector.AirzoneCloudConnector.get_installation', return_value=[])
    installation = AirzoneInstallation("1",email="j@j.com",password="1234")
    a = await installation.start()
    assert await get_installation_patch.called_once()


@pytest.mark.asyncio
async def test_installation_start_create_installation_devices(mocker):
    get_installation_patch = mocker.patch('airzonecloud.AirzoneCloudConnector.AirzoneCloudConnector.get_installation', return_value={"groups":[{"devices":[{"id":1},{"id":2}]}]})
    get_device_patch = mocker.patch('airzonecloud.AirzoneDevice.AirzoneDevice.get_device', return_value=[])
    installation = AirzoneInstallation("1",email="j@j.com",password="1234")
    a = await installation.start()
    assert get_device_patch.call_count ==2

@pytest.mark.asyncio
async def test_installation_start_if_device_is_system_the_installation_saves_it_as_master(mocker):
    get_installation_patch = mocker.patch('airzonecloud.AirzoneCloudConnector.AirzoneCloudConnector.get_installation', return_value={"groups":[{"devices":[{"device_id":1,"type":"az_system","ws_id":"1"}]}]})
    installation = AirzoneInstallation("1",email="j@j.com",password="1234")
    a = await installation.start()
    assert installation.system_device is not None
    assert isinstance(installation.system_device , AirzoneDevice_az_system)

@pytest.mark.asyncio
async def test_installation_append_installation_listener_to_the_connection(mocker):
    my_mock= AsyncMock()
    get_websocket_patch = my_mock.patch('airzonecloud.AirzoneCloudConnector.AirzoneCloudConnector.get_websocket')
    installation = AirzoneInstallation("1",email="j@j.com",password="1234")
    await installation.connect_live_updates()
    assert my_mock.called_once()

        



from .AirzoneCloudConnector import AirzoneCloudConnector
from .AirzoneDevice import AirzoneDevice, AirzoneDevice_az_system, AirzoneDevice_az_zone
import asyncio


class AirzoneInstallation:
    def __init__(self, installation_id, conn=None, email=None, password=None):
        if conn is None:
            conn = AirzoneCloudConnector(email, password)
        self._conn = conn
        self._id = installation_id
        self._inited = False
        self.devices = []
        self.system_device = None

    async def start(self):
        data = await self._conn.get_installation(self._id)
        if data:
            self._inited = True
            for group in data["groups"]:
                for device in group["devices"]:
                    dev = AirzoneDevice.get_device(device, self._conn, self)
                    self.devices.append(dev)
                    if isinstance(dev, AirzoneDevice_az_system):
                        self.system_device = dev

    async def connect_live_updates(self):
        ws_client = self._conn.get_websocket()
        ws_client.append_startup_command(ws_client.listen_instalation, (self._id,))
        asyncio.create_task(self._conn.get_websocket().open_client())

    def get_system_status(self):
        status = False
        for device in self.devices:
            if isinstance(device, AirzoneDevice_az_zone):
                status = status or device.status.get("power", False)

        return status
    
    def get_system_idle(self):
        status = False
        for device in self.devices:
            if isinstance(device, AirzoneDevice_az_zone):
                status = status or (
                    device.status.get("power", False)
                    and device.status.get("setpoint_air_heat", {"celsius": 0}).get(
                        "celsius", 0
                    )
                    > device.status.get("local_temperature", {"celsius": 0}).get(
                        "celsius", 0
                    )
                )

        return status


import asyncio
import json
import logging
import re
import websockets
from .const import *

logger = logging.getLogger("websockets")
logger.setLevel(logging.INFO)

class AirzoneWebsocketCLient:
    def __init__(self, host="m.airzonecloud.com") -> None:
        self._host = host
        self._ws = None
        self._connected = False
        self._token = None
        self._refresh_token = None
        self._next_command_id = 100
        self._startup_commands = []
        self._events = {}
        self._command_callbacks = {}
        self.add_command_callback("auth", self.authenticate)

    def set_token(self, token, refresh_token):
        self._token = token
        self._refresh_token = refresh_token

    def append_startup_command(self, call, args=None):
        self._startup_commands.append([call, args])

    def add_command_callback(self, command, method):
        if command not in self._command_callbacks:
            self._command_callbacks[command] = []
        self._command_callbacks[command].append(method)

    async def open_client(self):
        if self._token is None:
            raise Exception("Not authenticated")
        if self._connected:
            return
        while True:
            try:
                async with websockets.connect(
                    f"wss://{self._host}/api/v1/websockets/conn/?jwt={self._token}&transport=websocket&EIO=4",
                    origin=f"https://{self._host}",
                    extra_headers={
                        "Accept-Encoding": "gzip, deflate, br",
                        "Accept-Language": "es-419,es;q=0.9,en;q=0.8",
                        "Cache-Control": "no-cache",
                        "Pragma": "no-cache",
                        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.61 Safari/537.36",
                    },
                ) as websocket:
                    self._connected = True
                    self._ws = websocket

                    async for message in websocket:
                        await self.process_request(message, websocket)
            except BaseException as e:
                logger.error(e)

    async def process_request(self, message, websocket):

        op_code = re.findall(r"^\D*(\d+)", message[:2])[0]
        if op_code == OP_CODE_CONNECTED:
            await websocket.send(OP_CODE_CONNECTED_ACK)
        if op_code == OP_CODE_CONNECTED_ACK:
            await self.startup(websocket)
        elif op_code == OP_CODE_PING:
            await websocket.send(OP_CODE_PONG)
        elif op_code == OP_CODE_COMMAND:
            command_no = re.findall(r"^(\d*)", message[2:])[0]
            message_data = message[len(OP_CODE_COMMAND) + len(command_no) :]
            await self.response_to_command(command_no, json.loads(message_data))

        elif op_code == OP_CODE_COMMAND_RESPONSE:
            command_no = re.findall(r"^\D*(\d+)", message[2:])[0]
            message_data = message[len(OP_CODE_COMMAND) + len(command_no) :]
            await self.process_response(command_no, json.loads(message_data))

    async def startup(self, ws):
        for f in self._startup_commands:
            if f[1] is None:
                await f[0]()
            else:
                await f[0](*f[1])

    async def response_to_command(self, command_no, message_data):
        if message_data[0] in self._command_callbacks:
            for callback in self._command_callbacks[message_data[0]]:
                await callback(command_no, message_data[1], self)

        else:
            logger.info("commando no registrado:" + message_data[0])

    async def send_command(self, command, data=None):
        self._next_command_id += 1
        req_no = str(self._next_command_id)
        message = OP_CODE_COMMAND + str(req_no)
        if data is not None:
            message += json.dumps([command, data])
        else:
            message += json.dumps([command])
        event = asyncio.Event()
        self._events[req_no] = {"event": event, "data": None}
        await self._ws.send(message)
        a = await self.listen_to_response(req_no)
        return a

    async def clear_listeners(self):
        res = asyncio.create_task(self.send_command("clear_listeners"))

    async def authenticate(self, req_no, data, ws_client):
        if data == "authenticate":
            await self.send_response(req_no, [self._token])

    async def send_response(self, req_no, data):
        await self._ws.send(OP_CODE_COMMAND_RESPONSE + str(req_no) + json.dumps(data))

    async def listen_to_response(self, req_no):
        await self._events[req_no]["event"].wait()
        data = self._events[req_no]["data"]

        return data

    async def process_response(self, req_no, data):
        self._events[req_no]["data"] = data
        self._events[req_no]["event"].set()

    async def listen_instalation(self, installation_id):
        await asyncio.sleep(0.2)
        res = asyncio.create_task(
            self.send_command("listen_installation", installation_id)
        )
        return res
